# Generated from Scala.g4 by ANTLR 4.11.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .ScalaParser import ScalaParser
else:
    from ScalaParser import ScalaParser

# This class defines a complete generic visitor for a parse tree produced by ScalaParser.

class ScalaVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by ScalaParser#literal.
    def visitLiteral(self, ctx:ScalaParser.LiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#qualId.
    def visitQualId(self, ctx:ScalaParser.QualIdContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#ids.
    def visitIds(self, ctx:ScalaParser.IdsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#stableId.
    def visitStableId(self, ctx:ScalaParser.StableIdContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#classQualifier.
    def visitClassQualifier(self, ctx:ScalaParser.ClassQualifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#type_.
    def visitType_(self, ctx:ScalaParser.Type_Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#functionArgTypes.
    def visitFunctionArgTypes(self, ctx:ScalaParser.FunctionArgTypesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#existentialClause.
    def visitExistentialClause(self, ctx:ScalaParser.ExistentialClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#existentialDcl.
    def visitExistentialDcl(self, ctx:ScalaParser.ExistentialDclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#infixType.
    def visitInfixType(self, ctx:ScalaParser.InfixTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#compoundType.
    def visitCompoundType(self, ctx:ScalaParser.CompoundTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#annotType.
    def visitAnnotType(self, ctx:ScalaParser.AnnotTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#simpleType.
    def visitSimpleType(self, ctx:ScalaParser.SimpleTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#typeArgs.
    def visitTypeArgs(self, ctx:ScalaParser.TypeArgsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#types.
    def visitTypes(self, ctx:ScalaParser.TypesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#refinement.
    def visitRefinement(self, ctx:ScalaParser.RefinementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#refineStat.
    def visitRefineStat(self, ctx:ScalaParser.RefineStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#typePat.
    def visitTypePat(self, ctx:ScalaParser.TypePatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#ascription.
    def visitAscription(self, ctx:ScalaParser.AscriptionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#expr.
    def visitExpr(self, ctx:ScalaParser.ExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#expr1.
    def visitExpr1(self, ctx:ScalaParser.Expr1Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#prefixDef.
    def visitPrefixDef(self, ctx:ScalaParser.PrefixDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#postfixExpr.
    def visitPostfixExpr(self, ctx:ScalaParser.PostfixExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#infixExpr.
    def visitInfixExpr(self, ctx:ScalaParser.InfixExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#prefixExpr.
    def visitPrefixExpr(self, ctx:ScalaParser.PrefixExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#simpleExpr.
    def visitSimpleExpr(self, ctx:ScalaParser.SimpleExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#simpleExpr1.
    def visitSimpleExpr1(self, ctx:ScalaParser.SimpleExpr1Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#exprs.
    def visitExprs(self, ctx:ScalaParser.ExprsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#argumentExprs.
    def visitArgumentExprs(self, ctx:ScalaParser.ArgumentExprsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#args.
    def visitArgs(self, ctx:ScalaParser.ArgsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#blockExpr.
    def visitBlockExpr(self, ctx:ScalaParser.BlockExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#block.
    def visitBlock(self, ctx:ScalaParser.BlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#blockStat.
    def visitBlockStat(self, ctx:ScalaParser.BlockStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#resultExpr.
    def visitResultExpr(self, ctx:ScalaParser.ResultExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#enumerators.
    def visitEnumerators(self, ctx:ScalaParser.EnumeratorsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#generator.
    def visitGenerator(self, ctx:ScalaParser.GeneratorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#caseClauses.
    def visitCaseClauses(self, ctx:ScalaParser.CaseClausesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#caseClause.
    def visitCaseClause(self, ctx:ScalaParser.CaseClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#guard_.
    def visitGuard_(self, ctx:ScalaParser.Guard_Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#pattern.
    def visitPattern(self, ctx:ScalaParser.PatternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#pattern1.
    def visitPattern1(self, ctx:ScalaParser.Pattern1Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#pattern2.
    def visitPattern2(self, ctx:ScalaParser.Pattern2Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#pattern3.
    def visitPattern3(self, ctx:ScalaParser.Pattern3Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#simplePattern.
    def visitSimplePattern(self, ctx:ScalaParser.SimplePatternContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#patterns.
    def visitPatterns(self, ctx:ScalaParser.PatternsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#typeParamClause.
    def visitTypeParamClause(self, ctx:ScalaParser.TypeParamClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#funTypeParamClause.
    def visitFunTypeParamClause(self, ctx:ScalaParser.FunTypeParamClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#variantTypeParam.
    def visitVariantTypeParam(self, ctx:ScalaParser.VariantTypeParamContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#typeParam.
    def visitTypeParam(self, ctx:ScalaParser.TypeParamContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#paramClauses.
    def visitParamClauses(self, ctx:ScalaParser.ParamClausesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#paramClause.
    def visitParamClause(self, ctx:ScalaParser.ParamClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#params.
    def visitParams(self, ctx:ScalaParser.ParamsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#param.
    def visitParam(self, ctx:ScalaParser.ParamContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#paramType.
    def visitParamType(self, ctx:ScalaParser.ParamTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#classParamClauses.
    def visitClassParamClauses(self, ctx:ScalaParser.ClassParamClausesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#classParamClause.
    def visitClassParamClause(self, ctx:ScalaParser.ClassParamClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#classParams.
    def visitClassParams(self, ctx:ScalaParser.ClassParamsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#classParam.
    def visitClassParam(self, ctx:ScalaParser.ClassParamContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#bindings.
    def visitBindings(self, ctx:ScalaParser.BindingsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#binding.
    def visitBinding(self, ctx:ScalaParser.BindingContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#modifier.
    def visitModifier(self, ctx:ScalaParser.ModifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#localModifier.
    def visitLocalModifier(self, ctx:ScalaParser.LocalModifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#accessModifier.
    def visitAccessModifier(self, ctx:ScalaParser.AccessModifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#accessQualifier.
    def visitAccessQualifier(self, ctx:ScalaParser.AccessQualifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#annotation.
    def visitAnnotation(self, ctx:ScalaParser.AnnotationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#constrAnnotation.
    def visitConstrAnnotation(self, ctx:ScalaParser.ConstrAnnotationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#templateBody.
    def visitTemplateBody(self, ctx:ScalaParser.TemplateBodyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#templateStat.
    def visitTemplateStat(self, ctx:ScalaParser.TemplateStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#selfType.
    def visitSelfType(self, ctx:ScalaParser.SelfTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#import_.
    def visitImport_(self, ctx:ScalaParser.Import_Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#importExpr.
    def visitImportExpr(self, ctx:ScalaParser.ImportExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#importSelectors.
    def visitImportSelectors(self, ctx:ScalaParser.ImportSelectorsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#importSelector.
    def visitImportSelector(self, ctx:ScalaParser.ImportSelectorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#dcl.
    def visitDcl(self, ctx:ScalaParser.DclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#valDcl.
    def visitValDcl(self, ctx:ScalaParser.ValDclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#varDcl.
    def visitVarDcl(self, ctx:ScalaParser.VarDclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#funDcl.
    def visitFunDcl(self, ctx:ScalaParser.FunDclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#funSig.
    def visitFunSig(self, ctx:ScalaParser.FunSigContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#typeDcl.
    def visitTypeDcl(self, ctx:ScalaParser.TypeDclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#patVarDef.
    def visitPatVarDef(self, ctx:ScalaParser.PatVarDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#def_.
    def visitDef_(self, ctx:ScalaParser.Def_Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#patDef.
    def visitPatDef(self, ctx:ScalaParser.PatDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#varDef.
    def visitVarDef(self, ctx:ScalaParser.VarDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#funDef.
    def visitFunDef(self, ctx:ScalaParser.FunDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#typeDef.
    def visitTypeDef(self, ctx:ScalaParser.TypeDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#tmplDef.
    def visitTmplDef(self, ctx:ScalaParser.TmplDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#classDef.
    def visitClassDef(self, ctx:ScalaParser.ClassDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#traitDef.
    def visitTraitDef(self, ctx:ScalaParser.TraitDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#objectDef.
    def visitObjectDef(self, ctx:ScalaParser.ObjectDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#classTemplateOpt.
    def visitClassTemplateOpt(self, ctx:ScalaParser.ClassTemplateOptContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#traitTemplateOpt.
    def visitTraitTemplateOpt(self, ctx:ScalaParser.TraitTemplateOptContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#classTemplate.
    def visitClassTemplate(self, ctx:ScalaParser.ClassTemplateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#traitTemplate.
    def visitTraitTemplate(self, ctx:ScalaParser.TraitTemplateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#classParents.
    def visitClassParents(self, ctx:ScalaParser.ClassParentsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#traitParents.
    def visitTraitParents(self, ctx:ScalaParser.TraitParentsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#constr.
    def visitConstr(self, ctx:ScalaParser.ConstrContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#earlyDefs.
    def visitEarlyDefs(self, ctx:ScalaParser.EarlyDefsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#earlyDef.
    def visitEarlyDef(self, ctx:ScalaParser.EarlyDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#constrExpr.
    def visitConstrExpr(self, ctx:ScalaParser.ConstrExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#constrBlock.
    def visitConstrBlock(self, ctx:ScalaParser.ConstrBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#selfInvocation.
    def visitSelfInvocation(self, ctx:ScalaParser.SelfInvocationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#topStatSeq.
    def visitTopStatSeq(self, ctx:ScalaParser.TopStatSeqContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#topStat.
    def visitTopStat(self, ctx:ScalaParser.TopStatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#packaging.
    def visitPackaging(self, ctx:ScalaParser.PackagingContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#packageObject.
    def visitPackageObject(self, ctx:ScalaParser.PackageObjectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ScalaParser#compilationUnit.
    def visitCompilationUnit(self, ctx:ScalaParser.CompilationUnitContext):
        return self.visitChildren(ctx)



del ScalaParser


from dataclasses import dataclass
from dovado_rtl.parsers.utilities.antlr.antlr_module import AntlrModule
from dovado_rtl.parsers.utilities.port import Port


@dataclass
class ModuleWithPorts:
    ports: list[Port]


@dataclass
class HdlAntlrModule(AntlrModule, ModuleWithPorts):
    ...



from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Literal, Union

import toml
from dovado_rtl.explorers.utilities.range import Range

from dovado_rtl.parsers.utilities.parsed import (
    MODULE_NAME_STR,
    PARAMETER_NAME_STR,
    VALUE_STR,
)
import toml
import csv

SOURCE_PATH = Path


@dataclass
class Space(ABC):
    @abstractmethod
    def get_sources(self) -> list[SOURCE_PATH]:
        ...


@dataclass
class ContinuousSpace(Space):
    ranges: dict[SOURCE_PATH, dict[MODULE_NAME_STR, dict[PARAMETER_NAME_STR, Range]]]

    def __init__(self, toml_file: Path) -> None:
        try:
            toml_dict = toml.load(toml_file)
        except Exception as e:
            raise ValueError(
                "TOML file '" + str(toml_file) + " could not be loaded.\n" + str(e)
            )
        self.ranges = {}
        for source, module_parameter_dict in toml_dict.items():
            source_path = Path(source)
            self.ranges[source_path] = {}
            for module_name, parameter_range_dict in module_parameter_dict.items():
                self.ranges[source_path][module_name] = {}
                for parameter_name, range in parameter_range_dict.items():
                    self.ranges[source_path][module_name][parameter_name] = Range(range)

    def get_sources(self) -> list[SOURCE_PATH]:
        return list(self.ranges.keys())

    def get_structure(
        self,
    ) -> dict[SOURCE_PATH, dict[MODULE_NAME_STR, list[PARAMETER_NAME_STR]]]:
        out = {}
        for source, module_to_parameter in self.ranges.items():
            out[source] = {}
            for module, parameter_to_range in module_to_parameter.items():
                out[source][module] = []
                for parameter, _ in parameter_to_range.items():
                    out[source][module].append(parameter)
        return out

    def get_range(
        self,
        source_path: SOURCE_PATH,
        module_name: MODULE_NAME_STR,
        parameter_name: PARAMETER_NAME_STR,
    ) -> Range:
        return self.ranges[source_path][module_name][parameter_name]

    def get_ranges(self) -> list[tuple[PARAMETER_NAME_STR, Range]]:
        out = []
        for _, module_to_parameter in self.ranges.items():
            for _, parameter_to_range in module_to_parameter.items():
                for parameter, range in parameter_to_range.items():
                    out.append((parameter, range))
        return out

    def number_of_parameters(self) -> int:
        return len(self.get_ranges())

    def get_parameters(self) -> list[PARAMETER_NAME_STR]:
        return [parameter_range[0] for parameter_range in self.get_ranges()]

    def steps(self) -> list[Union[int, Literal["powers_of_two"]]]:
        return [range.step for _, range in self.get_ranges()]


@dataclass
class DiscreteSpace(Space):
    points: dict[
        SOURCE_PATH, dict[MODULE_NAME_STR, dict[PARAMETER_NAME_STR, list[VALUE_STR]]]
    ]

    def __init__(self, csv_file: Path, project_root: Path) -> None:
        try:
            csv_dict = self._strip_white_spaces(csv.DictReader(open(csv_file)))
        except Exception as e:
            raise ValueError(
                "CSV file '" + str(csv_file) + " could not be loaded.\n" + str(e)
            )
        fully_specified_parameters = list(csv_dict[0].keys())
        self.points = {}
        for fully_specified_parameter in fully_specified_parameters:
            source_path = self._find_source_path(
                fully_specified_parameter, project_root
            )
            if source_path not in self.points.keys():
                self.points[source_path] = {}

            module_name = self._find_module(fully_specified_parameter, project_root)
            if module_name not in self.points[source_path].keys():
                self.points[source_path][module_name] = {}

            parameter_name = self._find_parameter(fully_specified_parameter)
            self.points[source_path][module_name][parameter_name] = self._get_column(
                csv_dict, fully_specified_parameter
            )

    def get_sources(self) -> list[SOURCE_PATH]:
        return list(self.points.keys())

    def get_points(
        self,
        source_path: SOURCE_PATH,
        module_name: MODULE_NAME_STR,
        parameter_name: PARAMETER_NAME_STR,
    ) -> list[VALUE_STR]:
        return self.points[source_path][module_name][parameter_name]

    @staticmethod
    def _strip_white_spaces(csv_dict: csv.DictReader) -> list[Any]:
        return [{k.strip(): v.strip() for k, v in row.items()} for row in csv_dict]

    @staticmethod
    def _get_column(
        csv_dict: list[dict[str, Any]], fully_specified_parameter: str
    ) -> list[VALUE_STR]:
        return [row[fully_specified_parameter] for row in csv_dict]

    @staticmethod
    def _find_parameter(fully_specified_parameter: str) -> PARAMETER_NAME_STR:
        return fully_specified_parameter.split("/")[-1]

    @staticmethod
    def _find_module(
        fully_specified_parameter: str, project_root: Path
    ) -> MODULE_NAME_STR:
        source_path = DiscreteSpace._find_source_path(
            fully_specified_parameter, project_root
        )
        module_parameter_path = fully_specified_parameter.replace(str(source_path), "")

        # Module names such as module1/module2 mean that module2 is a submodule of module1
        # TODO: support this in the AntlrParsed
        return "/".join(module_parameter_path.split("/")[1:-1])

    @staticmethod
    def _find_source_path(fully_specified_parameter: str, project_root: Path) -> Path:
        accumulated_path = []
        for directory in fully_specified_parameter.split("/"):
            current_path = Path(project_root, *accumulated_path, directory)
            if current_path.is_file():
                return Path(*accumulated_path, directory)
            accumulated_path.append(directory)
        raise ValueError(
            "Could not find the source path for parameter "
            + str(fully_specified_parameter)
        )



from dataclasses import dataclass
from typing import Any, Literal, Union


@dataclass(frozen=True)
class Range:
    start: int
    end: int
    step: Union[int, Literal["powers_of_two"]]

    def __init__(self, raw_range: Union[list[int], dict[str, Any]]) -> None:
        if isinstance(raw_range, list):
            self._parse_range_list(raw_range)
            object.__setattr__(self, "step", 0)
        if isinstance(raw_range, dict):
            if "range" not in raw_range.keys():
                ValueError(
                    "When specifying a range as a dict please specify the 'range' key.\n"
                    + "Parsed dictionary is "
                    + str(raw_range)
                )
            self._parse_range_list(raw_range["range"])
            if "step" in raw_range.keys():
                step = raw_range["step"]
                if type(step) is int or step == "powers_of_two":
                    object.__setattr__(self, "step", raw_range["step"])
                else:
                    ValueError(
                        "A step may be specified either as an integer or as 'powers_of_two'."
                    )
            else:
                object.__setattr__(self, "step", 0)

    def _parse_range_list(self, range_list: list[int]) -> None:
        range_list_length = len(range_list)
        if range_list_length < 2 or range_list_length > 3:
            raise ValueError(
                "range list must contain two values (range start and end) and optionally a directive"
            )
        object.__setattr__(self, "start", range_list[0])
        object.__setattr__(self, "end", range_list[1])



from dovado_rtl.parsers.utilities.antlr.antlr_parsed import AntlrParsed


class CppParsed(AntlrParsed):
    @property
    def _parameter_initialization_prefix(self) -> str:
        raise ValueError(
            "In c++ we work only on defines, no concept of initialization exists"
        )



#!/usr/bin/env python3
from dataclasses import dataclass

from antlr4 import ParserRuleContext

from dovado_rtl.parsers.utilities.parameter import Parameter


@dataclass
class AntlrParameter(Parameter):
    rule: ParserRuleContext
    has_default: bool



from typing import Dict, Tuple
import pexpect
import re
import threading


class Vivado:
    def __init__(self) -> None:
        self.__vivado = pexpect.spawnu(
            "vivado -mode tcl -nolog -nojournal",
            encoding="utf-8",
            codec_errors="ignore",
        )
        self.__t_vivado_init = threading.Thread(
            target=self.__vivado.expect, args=("Vivado%",)
        )
        self.__t_vivado_init.start()

    def execute_command(self, command: str) -> str:
        if not self.__vivado.isalive():
            raise Exception(
                "Could not start Vivado, is it installed?"
                + " Is it in your PATH?\n"
                + " Try running 'vivado -mode tcl' yourself and"
                + " debug why it does not start"
            )
        if self.__t_vivado_init:
            print("Waiting for Vivado to start ...")
            self.__t_vivado_init.join()
            self.__vivado.sendline("set_param tcl.collectionResultDisplayLimit Inf")
            self.__vivado.expect("Vivado%", timeout=None)  # type: ignore
            self.__t_vivado_init = None

        self.__vivado.sendline(command)
        self.__vivado.expect("Vivado%", timeout=None)  # type: ignore
        return str(self.__vivado.before)

    def execute_script(self, tcl_script: str) -> Tuple[str, bool]:

        vivado_out = self.execute_command("source " + tcl_script)
        return vivado_out, not re.findall("ERROR:", vivado_out)

    def get_boards(self) -> Dict[str, str]:
        parts = [
            x.strip()
            for x in re.sub(
                "get_parts\r\n", "", self.execute_command("get_parts")
            ).split(" ")[1:]
        ]
        family = [
            x.strip()
            for x in re.sub(
                "[get_parts]\r\n",
                "",
                self.execute_command("get_property FAMILY [get_parts]"),
            ).split(" ")[3:]
        ]
        family[0] = family[0][len("[get_parts]\r\n") :]
        return dict(zip(parts, family))

    def is_board_valid(self, board: str) -> bool:
        boards = self.get_boards()
        if board in set(boards.keys()):
            return True
        return False



from dovado_rtl.explorers.utilities.spaces import (
    ContinuousSpace,
    DiscreteSpace,
    SOURCE_PATH,
)
from dovado_rtl.input import Input
from dovado_rtl.parsers.utilities.parsed import (
    MODULE_NAME_STR,
    PARAMETER_NAME_STR,
    VALUE_STR,
    Parsed,
)
from nacolla import ImmutableModel


class ParsedProject(Input):
    parameter_sources: dict[SOURCE_PATH, Parsed]
    target_source: Parsed

    def replace(
        self,
        source_path: SOURCE_PATH,
        module_parameter_to_value: dict[
            MODULE_NAME_STR, dict[PARAMETER_NAME_STR, VALUE_STR]
        ],
    ) -> str:
        return self.parameter_sources[source_path].replace(module_parameter_to_value)

    def get_parsed_project(self) -> "ParsedProject":
        # This method is meaningful only when called by a children
        parsed_project_attributes = ParsedProject.__fields__

        return ParsedProject(
            **{
                key: value
                for key, value in dict(self).items()
                if key in parsed_project_attributes
            }
        )


class AutomaticExplorationProject(ParsedProject):
    space: ContinuousSpace


class ManualExplorationProject(ParsedProject):
    space: DiscreteSpace


class EndExploration(ImmutableModel):
    ...



from dataclasses import dataclass
from typing import Literal, Union

PORT_DIRECTION = Union[Literal["input"], Literal["output"], Literal["inout"]]
PORT_DIMENSION = Union[Literal["scalar"], Literal["vectorial"]]


@dataclass
class Port:
    name: str
    direction: PORT_DIRECTION
    dimension: PORT_DIMENSION
    has_default: bool



from .port import Port
from .antlr.hdl.hdl_antlr_parser import HdlAntlrParser
from .antlr.hdl.hdl_antlr_module import HdlAntlrModule
from .antlr.antlr_parameter import AntlrParameter
from .port import PORT_DIRECTION, PORT_DIMENSION



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Union
from nacolla.stateful_callable import StatefulCallable
from dovado_rtl.explorers.utilities.tasks import ParsedProject, EndExploration
from dovado_rtl.explorers.utilities.design_points import (
    DesignPoint,
    EvaluatedDesignPoint,
)
import csv


class Explorer(StatefulCallable, ABC):
    _homonymous_files_limit = 100

    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def explore(self, task: ParsedProject) -> DesignPoint:
        ...

    @abstractmethod
    def update(
        self, evaluated_design_point: Union[EvaluatedDesignPoint, EndExploration]
    ) -> Union[DesignPoint, EndExploration]:
        ...

    @staticmethod
    def _get_file_name(work_dir: str, base_file_name: str) -> str:
        current_file = base_file_name + ".csv"
        if not Path(work_dir, current_file).exists():
            return current_file

        homonymous_files = 1
        while homonymous_files < Explorer._homonymous_files_limit:
            current_file = base_file_name + str(homonymous_files) + ".csv"
            if not Path(work_dir, current_file).exists():
                return current_file
            homonymous_files += 1

        raise Exception(
            "Found more than "
            + str(Explorer._homonymous_files_limit)
            + " homonymous files with "
            + str(base_file_name)
        )

    @staticmethod
    def _add_row_to_output(row: list[str], file: Path) -> None:
        opened_file = file.open("a", newline="")
        csv_writer = csv.writer(opened_file)
        csv_writer.writerow(row)
        opened_file.flush()
